import * as Utils from 'web3-utils';
import { ProvidersModuleFactory } from 'web3-providers';
import { formatters } from 'web3-core-helpers';
import { AbiCoder } from 'web3-eth-abi';
import { PromiEvent } from 'web3-core-promievent';
import isFunction from 'lodash/isFunction';
import isUndefined from 'lodash/isUndefined';
import cloneDeep from 'lodash/cloneDeep';
import isArray from 'lodash/isArray';
import { SendTransactionMethod, CallMethod, GetPastLogsMethod, EstimateGasMethod, MethodModuleFactory } from 'web3-core-method';
import { LogSubscription } from 'web3-core-subscriptions';
import { AbstractWeb3Module } from 'web3-core';

class AbiModel {
  constructor(mappedAbi) {
    this.abi = mappedAbi;
  }
  getMethod(name) {
    if (this.hasMethod(name)) {
      return this.abi.methods[name];
    }
    return false;
  }
  getEvent(name) {
    if (this.hasEvent(name)) {
      return this.abi.events[name];
    }
    return false;
  }
  getEvents() {
    return this.abi.events;
  }
  getEventBySignature(signature) {
    let event;
    Object.keys(this.abi.events).forEach(key => {
      if (this.abi.events[key].signature === signature) {
        event = this.abi.events[key];
      }
    });
    return event;
  }
  hasMethod(name) {
    return typeof this.abi.methods[name] !== 'undefined';
  }
  hasEvent(name) {
    return typeof this.abi.events[name] !== 'undefined';
  }
}

class AbiItemModel {
  constructor(abiItem) {
    this.abiItem = abiItem;
    this.signature = this.abiItem.signature;
    this.name = this.abiItem.name;
    this.anonymous = this.abiItem.anonymous;
    this.contractMethodParameters = [];
  }
  get requestType() {
    if (this.abiItem.type === 'function' || this.abiItem.type === 'constructor') {
      if (this.abiItem.constant === true) {
        return 'call';
      }
      return 'send';
    }
  }
  getInputLength() {
    if (isArray(this.abiItem.inputs)) {
      return this.abiItem.inputs.length;
    }
    return 0;
  }
  getInputs() {
    if (isArray(this.abiItem.inputs)) {
      return this.abiItem.inputs;
    }
    return [];
  }
  getOutputs() {
    if (isArray(this.abiItem.outputs)) {
      return this.abiItem.outputs;
    }
    return [];
  }
  givenParametersLengthIsValid() {
    const inputLength = this.getInputLength();
    if (this.contractMethodParameters.length === inputLength) {
      return true;
    }
    throw new Error(`The number of arguments is not matching the methods required number. You need to pass ${inputLength} arguments.`);
  }
  getIndexedInputs() {
    return this.getInputs().filter(input => {
      return input.indexed === true;
    });
  }
  isOfType(type) {
    return this.abiItem.type === type;
  }
}

class MethodEncoder {
  constructor(abiCoder) {
    this.abiCoder = abiCoder;
  }
  encode(abiItemModel, deployData) {
    let encodedParameters = this.abiCoder.encodeParameters(abiItemModel.getInputs(), abiItemModel.contractMethodParameters).replace('0x', '');
    if (abiItemModel.signature === 'constructor') {
      if (!deployData) {
        throw new Error('The contract has no contract data option set. This is necessary to append the constructor parameters.');
      }
      return deployData + encodedParameters;
    }
    if (abiItemModel.isOfType('function')) {
      return abiItemModel.signature + encodedParameters;
    }
    return encodedParameters;
  }
}

class EventFilterEncoder {
  constructor(abiCoder) {
    this.abiCoder = abiCoder;
  }
  encode(abiItemModel, filter) {
    const indexedInputs = abiItemModel.getIndexedInputs();
    let topics = [];
    indexedInputs.forEach(indexedInput => {
      if (filter[indexedInput.name]) {
        let filterItem = filter[indexedInput.name];
        if (isArray(filterItem)) {
          filterItem = filterItem.map(item => {
            return this.abiCoder.encodeParameter(indexedInput.type, item);
          });
          topics.push(filterItem);
          return;
        }
        topics.push(this.abiCoder.encodeParameter(indexedInput.type, filterItem));
      }
    });
    return topics;
  }
}

class AllEventsFilterEncoder extends EventFilterEncoder {
  encode(abiModel, filter) {
    const events = abiModel.getEvents();
    let topics = [];
    Object.keys(events).forEach(key => {
      topics.push(super.encode(events[key], filter));
    });
    return topics;
  }
}

class EventLogDecoder {
  constructor(abiCoder) {
    this.abiCoder = abiCoder;
  }
  decode(abiItemModel, response) {
    let argTopics = response.topics;
    if (abiItemModel.anonymous) {
      argTopics = response.topics.slice(1);
    }
    response.returnValues = this.abiCoder.decodeLog(abiItemModel.getInputs(), response.data, argTopics);
    response.event = abiItemModel.name;
    response.signature = abiItemModel.signature;
    response.raw = {
      data: response.data,
      topics: response.topics
    };
    if (abiItemModel.anonymous || !response.topics[0]) {
      response.signature = null;
    }
    delete response.data;
    delete response.topics;
    return response;
  }
}

class AllEventsLogDecoder extends EventLogDecoder {
  constructor(abiCoder) {
    super(abiCoder);
  }
  decode(abiModel, response) {
    return super.decode(abiModel.getEventBySignature(response.topics[0]), response);
  }
}

class AbiMapper {
  constructor(contractModuleFactory, abiCoder, utils) {
    this.utils = utils;
    this.abiCoder = abiCoder;
    this.contractModuleFactory = contractModuleFactory;
  }
  map(abi) {
    const mappedAbiItems = {
      methods: {},
      events: {}
    };
    abi.forEach(abiItem => {
      abiItem.constant = this.isConstant(abiItem);
      abiItem.payable = this.isPayable(abiItem);
      if (abiItem.name) {
        abiItem.funcName = this.utils.jsonInterfaceMethodToString(abiItem);
      }
      let abiItemModel;
      if (abiItem.type === 'function') {
        abiItem.signature = this.abiCoder.encodeFunctionSignature(abiItem.funcName);
        abiItemModel = this.contractModuleFactory.createAbiItemModel(abiItem);
        if (!mappedAbiItems.methods[abiItem.name]) {
          mappedAbiItems.methods[abiItem.name] = abiItemModel;
        } else {
          if (isArray(mappedAbiItems.methods[abiItem.name])) {
            mappedAbiItems.methods[abiItem.name].push(abiItemModel);
          } else {
            mappedAbiItems.methods[abiItem.name] = [mappedAbiItems.methods[abiItem.name], abiItemModel];
          }
        }
        mappedAbiItems.methods[abiItem.signature] = abiItemModel;
        mappedAbiItems.methods[abiItem.funcName] = abiItemModel;
        return;
      }
      if (abiItem.type === 'event') {
        abiItem.signature = this.abiCoder.encodeEventSignature(abiItem.funcName);
        abiItemModel = this.contractModuleFactory.createAbiItemModel(abiItem);
        if (!mappedAbiItems.events[abiItem.name] || mappedAbiItems.events[abiItem.name].name === 'bound ') {
          mappedAbiItems.events[abiItem.name] = abiItemModel;
        }
        mappedAbiItems.events[abiItem.signature] = abiItemModel;
        mappedAbiItems.events[abiItem.funcName] = abiItemModel;
      }
      if (abiItem.type === 'constructor') {
        abiItem.signature = abiItem.type;
        mappedAbiItems.methods['contractConstructor'] = this.contractModuleFactory.createAbiItemModel(abiItem);
      }
    });
    return this.contractModuleFactory.createAbiModel(mappedAbiItems);
  }
  isConstant(abiItem) {
    return abiItem.stateMutability === 'view' || abiItem.stateMutability === 'pure' || abiItem.constant;
  }
  isPayable(abiItem) {
    return abiItem.stateMutability === 'payable' || abiItem.payable;
  }
}

class MethodOptionsMapper {
  constructor(utils, formatters$$1) {
    this.utils = utils;
    this.formatters = formatters$$1;
  }
  map(contract, options) {
    let from = null;
    if (options.from) {
      from = this.utils.toChecksumAddress(this.formatters.inputAddressFormatter(options.from));
    }
    options.to = contract.address;
    options.from = from || contract.defaultAccount;
    options.gasPrice = options.gasPrice || contract.defaultGasPrice;
    options.gas = options.gas || options.gasLimit || contract.defaultGas;
    delete options.gasLimit;
    return options;
  }
}

class EventOptionsMapper {
  constructor(formatters$$1, eventFilterEncoder) {
    this.formatters = formatters$$1;
    this.eventFilterEncoder = eventFilterEncoder;
  }
  map(abiItemModel, contract, options) {
    if (!options) {
      options = {};
    }
    if (!isArray(options.topics)) {
      options.topics = [];
    }
    if (typeof options.fromBlock !== 'undefined') {
      options.fromBlock = this.formatters.inputBlockNumberFormatter(options.fromBlock);
    } else if (contract.defaultBlock !== null) {
      options.fromBlock = contract.defaultBlock;
    }
    if (typeof options.toBlock !== 'undefined') {
      options.toBlock = this.formatters.inputBlockNumberFormatter(options.toBlock);
    }
    if (!abiItemModel.anonymous) {
      options.topics.unshift(abiItemModel.signature);
    }
    if (typeof options.filter !== 'undefined') {
      options.topics = options.topics.concat(this.eventFilterEncoder.encode(abiItemModel, options.filter));
    }
    if (!options.address) {
      options.address = contract.address;
    }
    return options;
  }
}

class AllEventsOptionsMapper {
  constructor(formatters$$1, allEventsFilterEncoder) {
    this.formatters = formatters$$1;
    this.allEventsFilterEncoder = allEventsFilterEncoder;
  }
  map(abiModel, contract, options) {
    if (!options) {
      options = {};
    }
    options.topics = [];
    if (typeof options.fromBlock !== 'undefined') {
      options.fromBlock = this.formatters.inputBlockNumberFormatter(options.fromBlock);
    } else if (contract.defaultBlock !== null) {
      options.fromBlock = contract.defaultBlock;
    }
    if (typeof options.toBlock !== 'undefined') {
      options.toBlock = this.formatters.inputBlockNumberFormatter(options.toBlock);
    }
    if (typeof options.filter !== 'undefined') {
      options.topics = options.topics.concat(this.allEventsFilterEncoder.encode(abiModel, options.filter));
    }
    if (!options.address) {
      options.address = contract.address;
    }
    return options;
  }
}

class MethodsProxy {
  constructor(contract, abiModel, methodFactory, methodEncoder, methodOptionsValidator, methodOptionsMapper, PromiEvent$$1) {
    this.contract = contract;
    this.abiModel = abiModel;
    this.methodFactory = methodFactory;
    this.methodEncoder = methodEncoder;
    this.methodOptionsValidator = methodOptionsValidator;
    this.methodOptionsMapper = methodOptionsMapper;
    this.PromiEvent = PromiEvent$$1;
    return new Proxy(this, {
      get: (target, name) => {
        if (this.abiModel.hasMethod(name)) {
          let abiItemModel = this.abiModel.getMethod(name);
          let requestType = abiItemModel.requestType;
          if (isArray(abiItemModel)) {
            requestType = abiItemModel[0].requestType;
          }
          function anonymousFunction() {
            let methodArguments = [...arguments];
            if (!isArray(abiItemModel) && abiItemModel.isOfType('constructor')) {
              if (methodArguments[0]['data']) {
                target.contract.options.data = methodArguments[0]['data'];
              }
              if (methodArguments[0]['arguments']) {
                abiItemModel.contractMethodParameters = methodArguments[0]['arguments'];
              }
              return anonymousFunction;
            }
            if (isArray(abiItemModel)) {
              const abiItemModelFound = abiItemModel.some(model => {
                model.contractMethodParameters = methodArguments;
                try {
                  model.givenParametersLengthIsValid();
                } catch (error) {
                  return false;
                }
                abiItemModel = model;
                return true;
              });
              if (!abiItemModelFound) {
                throw new Error(`Methods with name "${name}" found but the given parameters are wrong`);
              }
              return anonymousFunction;
            }
            abiItemModel.contractMethodParameters = methodArguments;
            return anonymousFunction;
          }
          anonymousFunction[requestType] = function () {
            if (abiItemModel.isOfType('constructor')) {
              return target.executeMethod(abiItemModel, arguments, 'contract-deployment');
            }
            return target.executeMethod(abiItemModel, arguments, requestType);
          };
          anonymousFunction[requestType].request = function () {
            return target.createMethod(abiItemModel, arguments, requestType);
          };
          anonymousFunction.estimateGas = function () {
            return target.executeMethod(abiItemModel, arguments, 'estimate');
          };
          anonymousFunction.encodeABI = function () {
            return target.methodEncoder.encode(abiItemModel, target.contract.options.data);
          };
          return anonymousFunction;
        }
        if (target[name]) {
          return target[name];
        }
      }
    });
  }
  executeMethod(abiItemModel, methodArguments, requestType) {
    let method;
    try {
      method = this.createMethod(abiItemModel, methodArguments, requestType);
    } catch (error) {
      const promiEvent = new this.PromiEvent();
      method = this.methodFactory.createMethodByRequestType(abiItemModel, this.contract, requestType);
      method.arguments = methodArguments;
      promiEvent.reject(error);
      promiEvent.emit('error', error);
      if (isFunction(method.callback)) {
        method.callback(error, null);
      }
      return promiEvent;
    }
    if (requestType === 'call' || requestType === 'estimate') {
      return method.execute(this.contract);
    }
    return method.execute(this.contract, new this.PromiEvent());
  }
  createMethod(abiItemModel, methodArguments, requestType) {
    abiItemModel.givenParametersLengthIsValid();
    const method = this.methodFactory.createMethodByRequestType(abiItemModel, this.contract, requestType);
    method.arguments = methodArguments;
    if (typeof method.parameters[0] === 'undefined') {
      method.parameters[0] = {};
    }
    method.parameters[0]['data'] = this.methodEncoder.encode(abiItemModel, this.contract.options.data);
    method.parameters[0] = this.methodOptionsMapper.map(this.contract, method.parameters[0]);
    this.methodOptionsValidator.validate(abiItemModel, method);
    return method;
  }
}

class EventSubscriptionsProxy {
  constructor(contract, abiModel, eventSubscriptionFactory, eventOptionsMapper, eventLogDecoder, allEventsLogDecoder, allEventsOptionsMapper, PromiEvent$$1) {
    this.contract = contract;
    this.eventSubscriptionFactory = eventSubscriptionFactory;
    this.abiModel = abiModel;
    this.eventOptionsMapper = eventOptionsMapper;
    this.eventLogDecoder = eventLogDecoder;
    this.allEventsLogDecoder = allEventsLogDecoder;
    this.allEventsOptionsMapper = allEventsOptionsMapper;
    this.PromiEvent = PromiEvent$$1;
    return new Proxy(this, {
      get: (target, name) => {
        if (this.abiModel.hasEvent(name)) {
          return (options, callback) => {
            return target.subscribe(target.abiModel.getEvent(name), cloneDeep(options), callback);
          };
        }
        if (name === 'allEvents') {
          return (options, callback) => {
            return target.subscribeAll(cloneDeep(options), callback);
          };
        }
        return target[name];
      }
    });
  }
  subscribe(abiItemModel, options, callback) {
    if (options && !isUndefined(options.filter) && !isUndefined(options.topics)) {
      this.handleValidationError('Invalid subscription options: Only filter or topics are allowed and not both', callback);
      return;
    }
    return this.eventSubscriptionFactory.createEventLogSubscription(this.eventLogDecoder, this.contract, this.eventOptionsMapper.map(abiItemModel, this.contract, options), abiItemModel).subscribe(callback);
  }
  subscribeAll(options, callback) {
    if (options && !isUndefined(options.filter) && !isUndefined(options.topics)) {
      this.handleValidationError('Invalid subscription options: Only filter or topics are allowed and not both', callback);
      return;
    }
    return this.eventSubscriptionFactory.createAllEventsLogSubscription(this.allEventsLogDecoder, this.contract, this.allEventsOptionsMapper.map(this.abiModel, this.contract, options)).subscribe(callback);
  }
  handleValidationError(errorMessage, callback) {
    const error = new Error(errorMessage);
    if (isFunction(callback)) {
      callback(error, null);
    }
    throw error;
  }
}

class SendContractMethod extends SendTransactionMethod {
  constructor(utils, formatters$$1, transactionConfirmationWorkflow, accounts, transactionSigner, allEventsLogDecoder, abiModel) {
    super(utils, formatters$$1, transactionConfirmationWorkflow, accounts, transactionSigner);
    this.allEventsLogDecoder = allEventsLogDecoder;
    this.abiModel = abiModel;
  }
  afterExecution(response) {
    if (isArray(response.logs)) {
      response.events = {};
      response.logs.forEach((log, index) => {
        log = this.allEventsLogDecoder.decode(this.abiModel, log);
        if (log.event) {
          if (response.events[log.event]) {
            if (isArray(response.events[log.event])) {
              response.events[log.event].push(log);
              return;
            }
            response.events[log.event] = [response.events[log.event], log];
            return;
          }
          response.events[log.event] = log;
          return;
        }
        response.events[index] = log;
      });
      delete response.logs;
    }
    return response;
  }
}

class MethodOptionsValidator {
  constructor(utils) {
    this.utils = utils;
  }
  validate(abiItemModel, method) {
    if (!this.isToSet(abiItemModel, method)) {
      throw new Error("This contract object doesn't have address set yet, please set an address first.");
    }
    if (!this.isFromSet(method) && method instanceof SendContractMethod) {
      throw new Error('No valid "from" address specified in neither the given options, nor the default options.');
    }
    if (!this.isValueValid(abiItemModel, method)) {
      throw new Error('Can not send value to non-payable contract method or constructor');
    }
    return true;
  }
  isToSet(abiItemModel, method) {
    if (abiItemModel.signature === 'constructor') {
      return true;
    }
    return this.utils.isAddress(method.parameters[0].to);
  }
  isFromSet(method) {
    return this.utils.isAddress(method.parameters[0].from);
  }
  isValueValid(abiItemModel, method) {
    return !(!abiItemModel.payable && method.parameters[0].value && method.parameters[0].value > 0);
  }
}

class CallContractMethod extends CallMethod {
  constructor(utils, formatters$$1, abiCoder, abiItemModel) {
    super(utils, formatters$$1);
    this.abiCoder = abiCoder;
    this.abiItemModel = abiItemModel;
  }
  afterExecution(response) {
    if (!response) {
      return null;
    }
    if (response.length >= 2) {
      response = response.slice(2);
    }
    const result = this.abiCoder.decodeParameters(this.abiItemModel.getOutputs(), response);
    if (Object.keys(result).length === 1) {
      return result[0];
    }
    return result;
  }
}

class ContractDeployMethod extends SendTransactionMethod {
  constructor(utils, formatters$$1, transactionConfirmationWorkflow, accounts, transactionSigner, contract) {
    super(utils, formatters$$1, transactionConfirmationWorkflow, accounts, transactionSigner);
    this.contract = contract;
  }
  beforeExecution(moduleInstance) {
    super.beforeExecution(moduleInstance);
    delete this.parameters[0].to;
  }
  afterExecution(response) {
    const clonedContract = this.contract.clone();
    clonedContract.options.address = response.contractAddress;
    return clonedContract;
  }
}

class PastEventLogsMethod extends GetPastLogsMethod {
  constructor(utils, formatters$$1, eventLogDecoder, abiItemModel) {
    super(utils, formatters$$1);
    this.abiItemModel = abiItemModel;
    this.eventLogDecoder = eventLogDecoder;
  }
  afterExecution(response) {
    const formattedLogs = super.afterExecution(response);
    return formattedLogs.map(logItem => {
      return this.eventLogDecoder.decode(this.abiItemModel, logItem);
    });
  }
}

class MethodFactory {
  constructor(accounts, utils, formatters$$1, contractModuleFactory, methodModuleFactory, abiCoder) {
    this.accounts = accounts;
    this.utils = utils;
    this.formatters = formatters$$1;
    this.contractModuleFactory = contractModuleFactory;
    this.methodModuleFactory = methodModuleFactory;
    this.abiCoder = abiCoder;
  }
  createMethodByRequestType(abiItem, contract, requestType) {
    let rpcMethod;
    switch (requestType) {
      case 'call':
        rpcMethod = this.createCallContractMethod(abiItem);
        break;
      case 'send':
        rpcMethod = this.createSendContractMethod(abiItem, contract.abiModel);
        break;
      case 'estimate':
        rpcMethod = this.createEstimateGasMethod();
        break;
      case 'contract-deployment':
        rpcMethod = this.createContractDeployMethod(contract);
        break;
    }
    if (typeof rpcMethod === 'undefined') {
      throw new TypeError(`RPC call not found with requestType: "${requestType}"`);
    }
    return rpcMethod;
  }
  createPastEventLogsMethod(abiItem) {
    return new PastEventLogsMethod(this.utils, this.formatters, this.contractModuleFactory.createEventLogDecoder(), abiItem);
  }
  createCallContractMethod(abiItem) {
    return new CallContractMethod(this.utils, this.formatters, this.abiCoder, abiItem);
  }
  createSendContractMethod(abiItem, abiModel) {
    return new SendContractMethod(this.utils, this.formatters, this.methodModuleFactory.createTransactionConfirmationWorkflow(), this.accounts, this.methodModuleFactory.createTransactionSigner(), this.contractModuleFactory.createAllEventsLogDecoder(), abiModel);
  }
  createContractDeployMethod(contract) {
    return new ContractDeployMethod(this.utils, this.formatters, this.methodModuleFactory.createTransactionConfirmationWorkflow(), this.accounts, this.methodModuleFactory.createTransactionSigner(), contract);
  }
  createEstimateGasMethod() {
    return new EstimateGasMethod(this.utils, this.formatters);
  }
}

class EventLogSubscription extends LogSubscription {
  constructor(options, utils, formatters$$1, contract, getPastLogsMethod, eventLogDecoder, abiItemModel) {
    super(options, utils, formatters$$1, contract, getPastLogsMethod);
    this.eventLogDecoder = eventLogDecoder;
    this.abiItemModel = abiItemModel;
  }
  onNewSubscriptionItem(subscriptionItem) {
    return this.eventLogDecoder.decode(this.abiItemModel, this.formatters.outputLogFormatter(subscriptionItem));
  }
}

class AllEventsLogSubscription extends LogSubscription {
  constructor(options, utils, formatters$$1, contract, getPastLogsMethod, allEventsLogDecoder, abiModel) {
    super(options, utils, formatters$$1, contract, getPastLogsMethod);
    this.allEventsLogDecoder = allEventsLogDecoder;
    this.abiModel = abiModel;
  }
  onNewSubscriptionItem(subscriptionItem) {
    return this.allEventsLogDecoder.decode(this.abiModel, this.formatters.outputLogFormatter(subscriptionItem));
  }
}

class EventSubscriptionFactory {
  constructor(utils, formatters$$1) {
    this.utils = utils;
    this.formatters = formatters$$1;
  }
  createEventLogSubscription(eventLogDecoder, contract, options, abiItemModel) {
    return new EventLogSubscription(options, this.utils, this.formatters, contract, new GetPastLogsMethod(this.utils, this.formatters), eventLogDecoder, abiItemModel);
  }
  createAllEventsLogSubscription(allEventsLogDecoder, contract, options) {
    return new AllEventsLogSubscription(options, this.utils, this.formatters, contract, new GetPastLogsMethod(this.utils, this.formatters), allEventsLogDecoder, contract.abiModel);
  }
}

class AbstractContract extends AbstractWeb3Module {
  constructor(provider, providersModuleFactory, methodModuleFactory, contractModuleFactory, PromiEvent$$1, abiCoder, utils, formatters$$1, abi = AbstractWeb3Module.throwIfMissing('abi'), address, options = {}) {
    super(provider, providersModuleFactory, methodModuleFactory, null, options);
    this.contractModuleFactory = contractModuleFactory;
    this.abiCoder = abiCoder;
    this.utils = utils;
    this.formatters = formatters$$1;
    this.abiMapper = this.contractModuleFactory.createAbiMapper();
    this.options = options;
    this.PromiEvent = PromiEvent$$1;
    this.methodFactory = this.contractModuleFactory.createMethodFactory();
    this.abiModel = this.abiMapper.map(abi);
    this.options = options;
    if (address) {
      this.address = address;
    }
    this.methods = this.contractModuleFactory.createMethodsProxy(this, this.abiModel, this.PromiEvent);
    this.events = this.contractModuleFactory.createEventSubscriptionsProxy(this, this.abiModel, this.PromiEvent);
  }
  get jsonInterface() {
    return this.abiModel;
  }
  set jsonInterface(value) {
    this.abiModel = this.abiMapper.map(value);
    this.methods.abiModel = this.abiModel;
    this.events.abiModel = this.abiModel;
  }
  get address() {
    return this.options.address;
  }
  set address(value) {
    this.options.address = value;
  }
  once(eventName, options, callback) {
    if (!callback) {
      throw new Error('Once requires a callback function.');
    }
    if (options) {
      delete options.fromBlock;
    }
    const eventSubscription = this.events[eventName](options, callback);
    eventSubscription.on('data', () => {
      eventSubscription.unsubscribe();
    });
  }
  getPastEvents(eventName, options, callback) {
    return new Promise(async (resolve, reject) => {
      if (!this.abiModel.hasEvent(eventName)) {
        reject(new Error(`Event with name "${eventName}" does not exists.`));
      }
      const pastEventLogsMethod = this.methodFactory.createPastEventLogsMethod(this.abiModel.getEvent(eventName));
      pastEventLogsMethod.parameters = [options];
      pastEventLogsMethod.callback = callback;
      return resolve((await pastEventLogsMethod.execute(this)));
    });
  }
  deploy(options) {
    return this.methods.contractConstructor(options);
  }
  clone() {
    const contract = new this.constructor(this.currentProvider, this.providersModuleFactory, this.methodModuleFactory, this.contractModuleFactory, this.PromiEvent, this.abiCoder, this.utils, this.formatters, [], this.address, this.options);
    contract.abiModel = this.abiModel;
    contract.methods.abiModel = this.abiModel;
    contract.events.abiModel = this.abiModel;
    return contract;
  }
}

class ContractModuleFactory {
  constructor(utils, formatters$$1, abiCoder, accounts, methodModuleFactory) {
    this.utils = utils;
    this.formatters = formatters$$1;
    this.abiCoder = abiCoder;
    this.accounts = accounts;
    this.methodModuleFactory = methodModuleFactory;
  }
  createContract(provider, providersModuleFactory, PromiEvent$$1, abi, address, options) {
    return new AbstractContract(provider, providersModuleFactory, this.methodModuleFactory, this, PromiEvent$$1, this.abiCoder, this.utils, this.formatters, abi, address, options);
  }
  createAbiModel(mappedAbi) {
    return new AbiModel(mappedAbi);
  }
  createAbiItemModel(abiItem) {
    return new AbiItemModel(abiItem);
  }
  createMethodEncoder() {
    return new MethodEncoder(this.abiCoder);
  }
  createEventFilterEncoder() {
    return new EventFilterEncoder(this.abiCoder);
  }
  createAllEventsFilterEncoder() {
    return new AllEventsFilterEncoder(this.abiCoder);
  }
  createAbiMapper() {
    return new AbiMapper(this, this.abiCoder, this.utils);
  }
  createEventLogDecoder() {
    return new EventLogDecoder(this.abiCoder);
  }
  createAllEventsLogDecoder() {
    return new AllEventsLogDecoder(this.abiCoder);
  }
  createMethodOptionsValidator() {
    return new MethodOptionsValidator(this.utils);
  }
  createMethodOptionsMapper() {
    return new MethodOptionsMapper(this.utils, this.formatters);
  }
  createEventOptionsMapper() {
    return new EventOptionsMapper(this.formatters, this.createEventFilterEncoder());
  }
  createAllEventsOptionsMapper() {
    return new AllEventsOptionsMapper(this.formatters, this.createAllEventsFilterEncoder());
  }
  createMethodFactory() {
    return new MethodFactory(this.accounts, this.utils, this.formatters, this, this.methodModuleFactory, this.abiCoder);
  }
  createMethodsProxy(target, abiModel, PromiEvent$$1) {
    return new MethodsProxy(target, abiModel, this.createMethodFactory(), this.createMethodEncoder(), this.createMethodOptionsValidator(), this.createMethodOptionsMapper(), PromiEvent$$1);
  }
  createEventSubscriptionsProxy(contract, abiModel, PromiEvent$$1) {
    return new EventSubscriptionsProxy(contract, abiModel, this.createEventSubscriptionFactory(), this.createEventOptionsMapper(), this.createEventLogDecoder(), this.createAllEventsLogDecoder(), this.createAllEventsOptionsMapper(), PromiEvent$$1);
  }
  createEventSubscriptionFactory() {
    return new EventSubscriptionFactory(this.utils, this.formatters);
  }
}

const Contract = (provider, accounts, abi, address, options) => {
  return new ContractModuleFactory(Utils, formatters, new AbiCoder(), accounts, new MethodModuleFactory(accounts)).createContract(provider, new ProvidersModuleFactory(), PromiEvent, abi, address, options);
};

export { AbstractContract, ContractModuleFactory, Contract };
